from collections import defaultdict
import math
# Example query:
# dtl(
#    examples = [
#         {'Furniture': 'No', 'Nr. of rooms': '3', 'New kitchen': 'Yes', 'Acceptable': 'Yes'},
#         {'Furniture': 'Yes', 'Nr. of rooms': '3', 'New kitchen': 'No', 'Acceptable': 'No'},
#         {'Furniture': 'No', 'Nr. of rooms': '4', 'New kitchen': 'No', 'Acceptable': 'Yes'},
#         {'Furniture': 'No', 'Nr. of rooms': '3', 'New kitchen': 'No', 'Acceptable': 'No'},
#         {'Furniture': 'Yes', 'Nr. of rooms': '4', 'New kitchen': 'No', 'Acceptable': 'Yes'}
#    ],
#    attributes = {'Furniture': ['Yes', 'No'], 'Nr. of rooms': ['3', '4'], 'New kitchen': ['Yes', 'No'], 'Acceptable': ['Yes', 'No']},
#    target = 'Acceptable',
#    default = 'Yes'
# )
#
# Warning: the target attribute must not be used in the decision tree
# Warning: attributes are not necessarily binary
#
#
# Expected result:
# ('Nr. of rooms', {
#     '4': 'Yes',
#     '3': ('New kitchen', {
#         'Yes': 'Yes',
#         'No': 'No'}
#     )
#     }
# )

def calc_prob_given_target(examples, attributes, target):
    conditonal_probs = defaultdict(lambda: defaultdict(lambda: defaultdict(lambda: 0)))

    for tup in examples:
        for k, v in tup.items():
            if k != target:
                conditonal_probs[k][v][tup[target]] += 1


    for attr_name, value_dict in conditonal_probs.items():
        for attr_value , target_dict in value_dict.items():
            target_value_sum = sum(target_dict.values())
            for target_name, target_value in target_dict.items():
                conditonal_probs[attr_name][attr_value][target_name] = target_value / target_value_sum

    return conditonal_probs

def rank_attributes(examples, attributes, target):
    probs = defaultdict(lambda: defaultdict(lambda: 0))
    for example in examples:
        for k, v in example.items():
            probs[k][v] += 1
    for attr_name, dict in probs.items():
        values_sum = sum(dict.values())
        for k, v in dict.items():
            probs[attr_name][k] = v/values_sum


    target_entropy = 0

    for attr_name, dict in probs.items():
        if attr_name == target:
            for k, v in dict.items():
                target_entropy += - v * math.log2(v)

    conditonal_probs = calc_prob_given_target(examples, attributes, target)


    gain = defaultdict(lambda: 0)

    for attr_name, values_dict in attributes.items():
        conditional_entropy = 0
        if attr_name != target:
            for a in attributes[attr_name]:
                sum_of_conditional_entr = 0
                for val in attributes[target]:
                    if conditonal_probs[attr_name][a][val] != 0:
                        t = conditonal_probs[attr_name][a][val]
                        sum_of_conditional_entr += -1 * t * math.log2(t)
                conditional_entropy += probs[attr_name][a] * sum_of_conditional_entr

            gain[attr_name] = target_entropy - conditional_entropy

    gain = sorted(gain.items(), key=lambda item: -1*item[1])
    return gain


def check_all_examples_with_same_value_for_target(examples, target, default):
    all_target_values = []
    for example in examples:
        all_target_values.append(example[target])
    if len(all_target_values) == 0:
        return default, False
    return all_target_values.count(all_target_values[0]) == len(all_target_values), True

def build_tree_wrapper(examples, attributes, attribute_ranking, target, default):
    attr_name = attribute_ranking[0][0]
    head = (attr_name, {})
    flag, f = build_tree(examples, attributes, attribute_ranking, target, default)
    return f

def build_tree(examples, attributes, attribute_ranking, target, default, level=0, prev_value=None):
    v, flag =check_all_examples_with_same_value_for_target(examples, target, default)
    if v:
        if flag:
            t = [examples[0][attribute_ranking[level-1][0]], examples[0][target]]
        else:
            t = [prev_value, default]
        return True, t

    attr_name = attribute_ranking[level][0]
    head = (attr_name, {})
    for value in attributes[attr_name]:

        examples_with_value = [ example for example in examples if example[attr_name] == value]
        flag, v = build_tree(examples_with_value, attributes, attribute_ranking, target, default, level+1, prev_value=value)
        if flag:
            head[1][v[0]] = v[1]
        else:
            head[1][value] = v


    return False, head



def dtl(examples, attributes, target, default):


    attribute_ranking = rank_attributes(examples, attributes, target)
    tree = build_tree_wrapper(examples, attributes, attribute_ranking, target, default)


    return tree
